#!/usr/bin/env python
# coding: utf-8

"""
Although pyproject.toml is the new unified Python project settings file
that replaces setup.py, setup.py is needed for editable installations.
"""

from setuptools import setup

setup()
